#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
用于Q-learning的学习,内含MDP环境和Q-learning决策大脑
代码是根据莫凡Q-leaning教学视频修改而来
在这感谢莫凡的工作
附莫凡Q-learing学习网址：https://mofanpy.com/tutorials/machine-learning/reinforcement-learning/intro-q-learning/
我的GitHub地址：https://github.com/353055619/Q_learning_utils.git
"""

__author__ = 'Godw'

__all__ = ['env','Q_learning']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    Q-learning的MDP环境模块,示例如下：
    o----T
    -o---T
    --o--T
    ---o-T
    ----oT
    智能体'o'的目标是触碰到’T‘，
    'o'可以采用的动作值为['left', 'right']
    'o'触碰到‘T’，奖励1分；否则奖励0分
"""

__author__ = 'Godw'

import time
from typing import Tuple


class Env(object):
    ACTION = ['left', 'right']

    def __init__(self, length: int = 6, init_postion: int = 0, fresh_time: float = 0.3) -> None:
        """
        初始化
        :param length: 智能体完成任务需要走的最大步数
        :param init_postion: 智能体的初始位置
        :param fresh_time: 每步花费时间
        """
        self.l = length
        self.env_list = ['-'] * self.l + ['T']
        self.position = init_postion
        self.FRESH_TIME = fresh_time

    def refresh(self, init_postion: int = 0) -> None:
        """
        重置环境
        :param init_postion: 环境重置后，智能体‘o’所处的位置
        """
        self.position = init_postion

    def update_env(self, action: str) -> Tuple[int, int, bool]:
        """
        更新环境
        :param action: 采取的动作值
        :return: 本步奖励、下步位置、是否终止
        """
        reward = 0
        termination = False
        if action == 'right':
            self.position = self.position + 1
            if self.position == self.l:
                reward = 1
                termination = True
                self.position = self.position - 1
        if action == 'left':
            self.position = self.position - 1
            if self.position == -1:
                self.position = 0

        self.env_list[self.position] = 'o'
        print(''.join(self.env_list))
        time.sleep(self.FRESH_TIME)
        self.env_list = ['-'] * self.l + ['T']

        return reward, self.position, termination



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    Q-learning的决策模块
"""

__author__ = 'Godw'

from typing import List

import numpy as np
import pandas as pd
from pandas import DataFrame


def build_Q_Table(Action: List[str], Size: int) -> DataFrame:
    """
    创建一个Q表
    :param Action: 每步可以采取的动作列表
    :param Size: Q表的行数，也就是大小
    :return: Q表
    """
    table = pd.DataFrame(
        np.zeros((Size, len(Action))),
        columns=Action
    )
    return table


class Brain(object):
    def __init__(self, Q_Table: DataFrame, Action: List[str], Epsilon: float = 0.9, Alpha: float = 0.1,
                 Gamma: float = 0.9) -> None:
        """
        初始化
        :param Q_Table: Q表
        :param Action: 每步可以采取的动作列表
        :param Epsilon: ε
        :param Alpha: α
        :param Gamma: γ
        """
        self._Q_Table = Q_Table
        self._Action = Action
        self._Epsilon = Epsilon
        self._Alpha = Alpha
        self._Gamma = Gamma

    def choose_Action(self, Step: int) -> str:
        """
        决策下一步要采取的动作值
        :param Step: 当前步
        :return: 下一步采取的动作值
        """
        if np.random.uniform() > self._Epsilon or ((self._Q_Table.iloc[Step, :] == 0).all()):
            action = np.random.choice(self._Action)
        else:
            action = self._Action[np.argmax(self._Q_Table.iloc[Step, :])]
        return action

    def update_Q_Table(self, Action: str, Step: int, Reward: float, Step_: int, Done: bool
                       ) -> None:
        """
        更新Q表
        :param Action: 当前步采取的动作值
        :param Step: 当前步
        :param Reward: 执行当前步获得的奖励值
        :param Step_: 下一步
        :param Done: 是否结束
        """
        if Done:
            self._Q_Table.loc[Step, Action] = self._Q_Table.loc[Step, Action] + self._Alpha * (
                    Reward - self._Q_Table.loc[Step, Action])
        else:
            self._Q_Table.loc[Step, Action] = self._Q_Table.loc[Step, Action] + self._Alpha * (
                    Reward + self._Gamma * np.max(self._Q_Table.loc[Step_, Action]) - self._Q_Table.loc[Step, Action])
